# File: evm_AAA_Library.tcl
# Purpose: EVM Library
#

# $Id: evm_AAA_Library.tcl 303302 2017-11-09 14:23:45Z smarteau $

#Based on the following for Image Quality calculations
#
#L_0 = 46m
#D(VLT) = 8.2m, D(VISTA) = 4.0m, D(VST) = 2.5m

#VLT case:
#F_Kolb = 1/(1+300*D/L_0)-1 = 1/(1+300*8.2/46)-1 = -0.99
#VST case: F_Kolb = -0.97, VISTA case: F_Kolb = -0.98
#
#the variable instr_FK_ITF(instrument) {telescope F_Kolb ITF} is defined here:

set instr_FK_ITF(CRIRES) { "VLT" -0.99 0.1 }
set instr_FK_ITF(FLAMES) { "VLT" -0.99 0.00 }
set instr_FK_ITF(FORS_HR) { "VLT" -0.99 0.04 }
set instr_FK_ITF(FORS_SR) { "VLT" -0.99 0.05 }
set instr_FK_ITF(FORS2) { "VLT" -0.99 0.05 }
set instr_FK_ITF(HAWKI) { "VLT" -0.99 0.20 }
set instr_FK_ITF(KMOS) { "VLT" -0.99 0.23 }
set instr_FK_ITF(MUSE) { "VLT" -0.99 0.00 }
set instr_FK_ITF(OMEGACAM) {"VST" -0.97 0.35 }
set instr_FK_ITF(SINFONI) { "VLT" -0.99 0.01 } ; # noAO only
set instr_FK_ITF(UVES) { "VLT" -0.99 0.18 }
set instr_FK_ITF(VIMOS) { "VLT" -0.99 0.25 }
set instr_FK_ITF(VIRCAM) { "VISTA" -0.98 0.5 }
set instr_FK_ITF(VISIR) { "VLT" -0.99 0.01 }
set instr_FK_ITF(XSHOOTER) { "VLT" -0.99 0.40 }

####################################################################
##
## Parsing of OBs and Containers
##
####################################################################

# getParamValue
#
# returns value of a keyword in a given template, or a
# default value if not found
#
proc getParamValue { template keyword default_value } {

    # Get value of keyword in the template
    set value [evmParameterValue $template $keyword]

    # If value is empty (keyword not found)
    if { [string length $value] == 0 } {

	# Set the result to the specified default value
	set value $default_value

    }

    # Return result
    return $value

}

# Procedure: parseOB
#
#
proc parseOB { ob } {

    global OBdata scriptingBox

    set ob_id [evmID $ob]

    # List of all templates (Java handles)
    set templates [evmTemplates $ob]

    # template counter (acquisition=0)
    set index 0

    # Record OB type and name
    set OBdata($ob_id,type) [evmType $ob]
    set OBdata($ob_id,name) [evmName $ob]

    # Record observing mode (visitor/service)
    set OBdata($ob_id,runMode) [evmRunMode $ob]

    # Record target info (not for calibration blocks)
    if { $OBdata($ob_id,type) != "C" } {
	set target [evmTarget $ob]
	set OBdata($ob_id,targ_name) [evmName $target]
	set OBdata($ob_id,ra) [evmRightAscension $target]
	set OBdata($ob_id,dec) [evmDeclination $target]
	set OBdata($ob_id,ra_marcsec) [evmRightAscension $target 1]
	set OBdata($ob_id,dec_marcsec) [evmDeclination $target 1]
	set OBdata($ob_id,pm_ra) [evmPropRA $target]
	set OBdata($ob_id,pm_dec) [evmPropDec $target]
	set OBdata($ob_id,diff_ra) [evmDiffRA $target]
	set OBdata($ob_id,diff_dec) [evmDiffDec $target]
	set OBdata($ob_id,equinox) [evmEquinox $target]
	set OBdata($ob_id,epoch) [evmEpoch $target]

	# Get constraints from constraint set (not for calibration blocks)
	set constraintSet [evmConstraintSet $ob]
	set OBdata($ob_id,csAirmass) [evmAirMass $constraintSet]
	set OBdata($ob_id,csSeeing) [evmSeeing $constraintSet]
	set OBdata($ob_id,csTransp) [evmSkyTransparency $constraintSet]
	set OBdata($ob_id,csFLI) [evmFractionalLunarIllumination $constraintSet]
	set OBdata($ob_id,csMoonDist) [evmMoonAngularDistance $constraintSet]
	set OBdata($ob_id,csTwilight) [evmTwilight $constraintSet]
	set OBdata($ob_id,csPWV) [evmWatervapour $constraintSet]
	set OBdata($ob_id,csStrehl) [evmStrehlRatio $constraintSet]
	set OBdata($ob_id,csATM) [evmAtm $constraintSet]

	set OBdata($ob_id,csBaseline) [evmBaseline $constraintSet]

	# User Comments
	set OBdata($ob_id,userComments) [evmUserComments $ob]

	# Instrument Comments
	set OBdata($ob_id,instComments) [evmInstrumentSpecificComments $ob]

	# LST Time Intervals
	set timeIntervals [evmSTTimeIntervals $ob]

	set ti_idx 1
	foreach int $timeIntervals {
	    set OBdata($ob_id,LST,$ti_idx,start) [evmStartTime $int]
	    set OBdata($ob_id,LST,$ti_idx,end) [evmEndTime $int]
	    incr ti_idx
	}

	set OBdata($ob_id,LST,nbIntervals) [expr $ti_idx - 1]

    }

    # Loop on all templates
    foreach temp $templates {

	# Parse all valid templates (verification required to
	# catch exception for 1st template of CalBlocks which is always NULL)
	if { [evmValid $temp] } {

	    # Record template name
	    set OBdata($ob_id,$index,tplName) [evmName $temp]

	    # Parse individual template
	    parseTemplate $temp $ob_id $index

	    incr index ; # template counter: add 1
	}

    }

    set OBdata($ob_id,nbTpl) $index

    return
}

# Procedure: parseTemplate
# arguments:
#  - template: EVM template object
#  - ob_id : OB ID
#  - idx : index of the template inside the OB
#  - keywordList (global variable): list of template keywords to extract
#
# returns: nothing (global array is used)
#
# Parses a template and puts the parameters into
# a global array
#
proc parseTemplate { template ob_id idx } {

    # Global array in which parameters are stored
    global OBdata keywordList

    # Loop on all keywords to retrieve
    foreach {key def} $keywordList {

	# Retrieve value of keyword from template, assign default value if not found
	set OBdata($ob_id,$idx,$key) [getParamValue $template $key $def]

    }

    return
}

# Procedure: retrieveAndParse

#
# Retrieves the IDs and Java handles for all containers and OBs
# pertaining to the user's selection in P2PP. The data is stored
# in global variables.
# Each considered OB is also parsed into a global array.
#
# EXAMPLE: with one OB selected within a 3-OB container and one loose OB,
# like depicted below:
#
# +--+ TIME_LINK: id=119 / Name=Matter of time
# |  |
# |  +-- OB: id=121 / Name=the beginning
# |  +-- OB: id=124 / Name=the follow-up  SELECTED
# |  +-- OB: id=127 / Name=the end
# |
# +--+ OB: id=134 / Name=OB at the bottom  SELECTED
#
#
# the compound list will look like this;
#
# {{TIME_LINK 119 java0x27 0} {{OB 121 java0x28 1} {OB 124 java0x29 0} {OB 127 java0x2a 1}}} {OB 134 java0x3b}
#
proc retrieveAndParse {} {

    global scriptingBox compoundList compoundListObjects OBdata

    set parsedOB_ids {}
    set parsedContainer_ids {}
    set selectedOBList {}
    set selectedOB_ids {}
    set selectedContainerList {}
    set selectedContainer_ids {}

    set compoundList {}
    set compoundListObjects {}
    
    # First we get the list of all selected OBs from the API
    # Get total number of selected OBs
    set numOBs [evmGetOBCount]
    
    set obNum 0
    while { $obNum < $numOBs } {
	    
	set ob [evmGetOB $obNum]
	set obID [evmID $ob]
	
	lappend selectedOBList [list $ob $obID]
	lappend selectedOB_ids $obID
	incr obNum
    }
    
    # Same thing with the list of selected containers	
    # Get total number of OB containers present in the selection
    # Note that we deliberately skip 'folders' because they are without
    # any operational meaning.
    set numContainers [evmGetContainerCount]
    
    set containerNum 0
    while { $containerNum < $numContainers } {
	
	set container [evmGetContainer $containerNum]
	
	# Handling of exceptions:
	#  - folders are simply excluded
	#  - empty containers are excluded
	if { [evmType $container] == "FOLDER" || [$container getNumObsBlocks] == 0} {incr containerNum ; continue}

	set containerID [evmID $container]
	lappend selectedContainerList [list $container $containerID]
	lappend selectedContainer_ids $containerID

	incr containerNum
    }

    # Part I. Loop on selected OBs
    
    foreach obEntry $selectedOBList {

	set ob [lindex $obEntry 0]
	set obID [lindex $obEntry 1]
	if { [lsearch -exact $parsedOB_ids $obID] != -1 } { continue }
	
	# Is this OB part of a container?
	set parent [$ob getParentContainer]

	if { ![evmValid $parent] } {

	    # No parent? This is a loose OB. Parse it, store it.
	    # Parse it unless we have done it before
	    if { [lsearch -exact $parsedOB_ids $obID] == -1 } {
		lappend parsedOB_ids $obID
		parseOB $ob
	    }

	    lappend compoundList [list OB $obID $ob 1]
	    lappend compoundListObjects [list OB $ob 1]
	    
	} else {

	    # OB is in a container. We parse the entire container.
	    set containerList {}
	    set containerListObjects {}
	    
	    # Collect info about container and store in list
	    set containerID [evmID $parent]
	    set containerType [$parent getType]
	    set containerSelected 0
	    if { [lsearch -exact $selectedContainer_ids $containerID] != -1 } {
		set containerSelected 1
	    }
	    set containerList [list $containerType $containerID $parent $containerSelected]
	    set containerListObjects [list $containerType $parent $containerSelected]
	    
	    # How many OBs does the container have
	    set nbOBs [$parent getNumObsBlocks]

	    set obList {}
	    set obListObjects {}
	    
	    # Loop through all OBs of the container
	    for {set i 0} {$i < $nbOBs} {incr i} {

		# get OB #i (Java pointer)
		set ob [$parent getObsBlock $i]

		# get OB ID
		set obID [evmID $ob]

		set obSelected 0
		if { [lsearch -exact $selectedOB_ids $obID] != -1 } {
		    set obSelected 1
		}
		
		# Parse the OB unless we have done it before
		if { [lsearch -exact $parsedOB_ids $obID] == -1 } {
		    lappend parsedOB_ids $obID
		    parseOB $ob
		}

		lappend obList [list OB $obID $ob $obSelected]
		lappend obListObjects [list OB $ob $obSelected]

	    }
	    
	    lappend parsedContainer_ids $containerID
	    
	    lappend compoundList [list $containerList $obList]
	    lappend compoundListObjects [list $containerListObjects $obListObjects]
	    
	}
	
    } ; # end of loop on selected OBs

    # Part II. Loop on selected containers

    foreach containerEntry $selectedContainerList {

	set container [lindex $containerEntry 0]
	set containerID [lindex $containerEntry 1]
	if { [lsearch -exact $parsedContainer_ids $containerID] != -1 } { continue }
	
	set containerType [evmType $container]
	set containerSelected 1
	
	set containerList [list $containerType $containerID $container $containerSelected]
	set containerListObjects [list $containerType $container $containerSelected]
	    
	# How many OBs does the container have
	set nbOBs [$container getNumObsBlocks]

	set obList {}
	set obListObjects {}
	
	# Loop through all OBs of the container
	for {set i 0} {$i < $nbOBs} {incr i} {
	    
	    # get OB #i (Java pointer)
	    set ob [$container getObsBlock $i]
	    
	    # get OB ID
	    set obID [evmID $ob]
	    
	    set obSelected 0
	    if { [lsearch -exact $selectedOB_ids $obID] != -1 } {
		set obSelected 1
	    }
	    
	    # Parse the OB unless we have done it before
	    if { [lsearch -exact $parsedOB_ids $obID] == -1 } {
		lappend parsedOB_ids $obID
		parseOB $ob
	    }
	    
	    lappend obList [list OB $obID $ob $obSelected]
	    lappend obListObjects [list OB $ob $obSelected]
	    
	}
		
	lappend parsedContainer_ids $containerID
	
	lappend compoundList [list $containerList $obList]
	lappend compoundListObjects [list $containerListObjects $obListObjects]
	
    } ; # end of loop on selected containers

    # We store the total number of OBs we parsed
    # (independently of whether they are explicitly selected or not)
    set OBdata(nbOBs) [llength $parsedOB_ids]
	
}



# collectPhase1Info
#
# Retrieves Phase 1 information for later use within the main
# test procedure.
#
# input: none
# output: none (uses global array)
#
proc collectPhase1Info {} {

    global Phase1data compoundList

    # We use the first OB in the overall list to get a proper handle
    # for the EVM functions

    set object [lindex $compoundList 0]
    set objType [lindex $object 0]
    if { [llength $objType] > 1 } {

	set listOfObsBlocks [lindex $object 1]
	set objRef [lindex [lindex $listOfObsBlocks 0] 2] ; # Object reference (Java handle)

    } else {

	set objRef [lindex $object 2] ; # Object reference (Java handle)
    }

    set runID [evmRunID $objRef]

    set phase1Targets [evmPhase1Targets $objRef]

    set radecList {}

    set tIdx 1

    foreach t $phase1Targets {

	set ra [evmPhase1Ra $t ]
	set dec [evmPhase1Dec $t ]

	set radecString ""
	append radecString $ra [string trim $dec " +"]

	if { [set item [lsearch -glob $radecList [concat "$radecString" *]]] != -1 } {

	    set idx [lindex [lindex $radecList $item] 1]
	    append Phase1data(target,$idx,runRefs) [evmPhase1RunReferences $t]

	} else {

	    lappend radecList [concat "$radecString" $tIdx]

	    set Phase1data(target,$tIdx,RA) $ra
	    set Phase1data(target,$tIdx,DEC) $dec
	    set Phase1data(target,$tIdx,name) [evmPhase1TargetName $t]
	    set Phase1data(target,$tIdx,RA_arcsec) [expr [evmPhase1Ra $t 1] / 1000.]
	    set Phase1data(target,$tIdx,DEC_arcsec) [expr [evmPhase1Dec $t 1] / 1000.]
	    set Phase1data(target,$tIdx,runRefs) [evmPhase1RunReferences $t]
	    incr tIdx

	}

    }

    set Phase1data(target,nbTargets) [expr $tIdx - 1]

    set Phase1data(runFlags) [evmRunFlags $objRef]
    set Phase1data(tooFlag) [string range $Phase1data(runFlags) 2 2]
    # uncomment the 3 lines below for testing
    #evmInfo "\n\n\n\n\n\n*************** TOO FLAG SET TO TRUE FOR TESTING *****************\n\n"
    #evmInfo "*************** Please let me know after you are finished with testing (with TOO true)\n               so that you may continue testing with the hard-coded flag removed. *****************\n\n\n\n\n\n"
    #set Phase1data(tooFlag) 1

    set Phase1data(allocatedTime) [evmRunNights $objRef]

    set Phase1data(seeing) [evmRunSeeing $objRef]

    if { $Phase1data(seeing) == "n" } { set Phase1data(seeing) 2.0 }
    if { $Phase1data(seeing) == "" } { set Phase1data(seeing) 2.0 }
    if { $Phase1data(seeing) == "null" } { set Phase1data(seeing) 2.0 }

    set Phase1data(skyTransp) [evmRunSkyTransp $objRef]

    set Phase1data(moonFLI) [evmRunMoon $objRef]

    set Phase1data(runMode) [evmRunMode $objRef]

    # Specific to OmegaCAM: identify public surveys
    if { [evmInstrumentName $objRef] == "OMEGACAM" } {
	if { [regexp {^7730} $runID] == 1 } {
	    set Phase1data(isPublicSurvey) 1
	} else {
	    set Phase1data(isPublicSurvey) 0
	}
    }

    # Specific to VIRCAM: identify public surveys
    if { [evmInstrumentName $objRef] == "VIRCAM" } {
	if { [regexp {^7920} $runID] == 1 || [regexp {^9820} $runID] == 1} {
	    set Phase1data(isPublicSurvey) 1
	} else {
	    set Phase1data(isPublicSurvey) 0
	}
    }

}

####################################################################
##
## Auxiliary procedures for common checks
##
####################################################################

# checkTarget
#
#
# Procedure: check target position vs airmass
# report WARNING if target never above airmass 1.5 or 2.0
# report ERROR if target never meets airmass constraint
#
# if calOB, SKIP
#
# inputs:
#   - OBid: local OB ID
#
# output:
#   none. Issues warning/error messages if necessary
#
proc checkTarget { OBid runID {isTelUT 1} } {

    global OBdata Phase1data

    # Latitude of VLT Observatory (Paranal = -24:27:30)
    set latitude -24.63

    # Conversion factor rad => deg
    set DegreePerRadian 57.29578

    # Value of PI
    set PI  3.14159

    set targetOK 1

    # Get OB type (Obs/Calib)
    set obType $OBdata($OBid,type)

    # Set observing mode (visitor/service)
    set obRunMode $OBdata($OBid,runMode)

    # if calOB, return 1 (no target so always "OK")
    if { $obType == "C" } { return 1 }

    ### Target Name CHECKS ###

    #1. [General Check] Check Target name for unallowed characters:
    set targName $OBdata($OBid,targ_name)

    if { [regexp {^[a-zA-Z0-9+_ \.-]+$} $targName] == 0 } {

	set msg "The target name contains forbidden characters (allowed characters: letters, digits, space, + - _ . )."
	issueError $OBid $msg $runID

    }

    ### Target Coordinates CHECKS ###

    set targAlpha $OBdata($OBid,ra_marcsec)
    set targDelta $OBdata($OBid,dec_marcsec)

    #1. [General Check] Null coordinates ==> ToO OB? (send warning)
    if { ($targAlpha == 0.0 && $targDelta == 0.0)  && $Phase1data(tooFlag) != 1 } {

	issueWarning $OBid "This is not a ToO run, please check the target coordinates."

    }

    ### Airmass constraints CHECKS ###

    #1. [General Check] Airmass constraint: is constraint is > 2.9 ?
    # Get airmass from constraint set
    set csAirmass $OBdata($OBid,csAirmass)
    if { $obRunMode == "s" && $csAirmass >= 2.9 } {
	# issue an ERROR message and set flag.
	issueError $OBid "The airmass constraint must  be smaller than 2.9 due to the physical limit of the telescope." $runID
	set targetOK 0
    }

    #2. [General Check] Target position vs airmass:
    # Compute minimum zenithal distance
    set dec [expr $OBdata($OBid,dec_marcsec) / 3600000. ]
    set zenthDistance [expr abs($latitude - $dec)]

    #2b. for Service Mode runs: does object ever rise above csAirmass?
    if { $obRunMode == "s" } {

	# Get minimum target airmass
	set targAirmass [expr (1. / cos($zenthDistance/$DegreePerRadian) ) ]

	# If target minimum airmass > constraint airmass,
	if { $targAirmass > $csAirmass } {

	    # issue an ERROR message and set flag.
	    issueError $OBid "the target never rises above the chosen airmass constraint (min airmass: [format %.1f $targAirmass])" $runID
	    set targetOK 0
	}
    }

    return $targetOK
}

# checkEpoch [General Check]
#
#
# Procedure: check validity of epoch
#
#
# inputs:
#   - OBid: local OB ID
#   - runID: run ID
#
# output:
#   none. Issues error messages if necessary
#
proc checkEpoch { OBid runID } {

    global OBdata
    # Set observing mode (visitor/service)
    set obRunMode $OBdata($OBid,runMode)

    set epoch $OBdata($OBid,epoch)
    if { $epoch < 1950.0 || 2050.0 < $epoch } {
	set msg "Epoch $epoch is not valid; should be in the range 1950 - 2050."
	if { $obRunMode == "s" } {
	    issueError $OBid $msg $runID
	} else {
	    issueWarning $OBid $msg
	}
    }
}

#End checkEpoch

# checkProperMotions
#
#
# issues a warning if proper motions are larger
# than some predefined value, to try and prevent
# erroneous entries
#
# input: target package
#
proc checkProperMotions { ob_id } {

    global OBdata

    # Proper motions are expressed in arcsec/year
    # in the target package
    set pma $OBdata($ob_id,pm_ra)
    set pmd $OBdata($ob_id,pm_dec)

    if { [max [expr abs($pma)] [expr abs($pmd)]] > 1.0 } {

	set msg "At least one of the proper motions you specified is larger than 1 arcsec/year.\n         Please check that you used the correct unit of arcsec/year."
	issueWarning $ob_id $msg

    }

}

# [General Check]:
# Procedure: checkEphemerisFile
# Checks the format/contents of an ephemeris file
#
# input:
#   - ob_id : ID of the OB to which the ephemeris file is attached.
#   - runID : ID of the observing run to which the OB belongs
#   - ephData : ephemeris data as provided by the API function 'evmEFData'
#   - timeIntervals : absolute time intervals attached to the OB, as provided
#                     by 'evmTimeIntervals'
#
# output: none (all checks performed in the procedure)
#
proc checkEphemerisFile { ob_id runID ephData timeIntervals } {
    global OBdata

    # Set observing mode (visitor/service)
    set obRunMode $OBdata($ob_id,runMode)

    if { $obRunMode == "s" } {

	# Split contents of ephemeris file into individual lines
	set ephLines [split $ephData "\n"]

	# Make a single list of the time intervals of the OB
	set timeIntList {}
	set dateList {}

	foreach ti $timeIntervals {
	    lappend timeIntList [list [evmStartTime $ti] [evmEndTime $ti]]

	    set t1 [evmStartTime $ti] ; set t2 [evmEndTime $ti]
	    for {set t $t1} {$t <= $t2} {incr t 86400} {
		lappend dateList [clock format $t -format "%Y-%m-%d"]
	    }
	    if {$t < $t2 && [expr $t2 - $t1] > 86400} {
		lappend dateList [clock format $t2 -format "%Y-%m-%d"]
	    }
	}

	# Variable init
	set lineCount 0 ; # count lines
	set recordCount 0 ; # count valid ephemeris records
	set PAFMarkerCount 0 ; # count PAF header start/end

	# Record data of previous line for coord comparison
	set previousDateTime ""
	set previousRAarcsec ""
	set previousDECarcsec ""

	## Loop on lines of the ephemeris file
	foreach line $ephLines {

	    incr lineCount
	    set value ""

	    # If the line does not contain an ephemeris record
	    if { ![regexp {^INS\.EPHEM\.RECORD\ +\"(.*)\"} $line dummy value] } {

		# Check to see if the file has a PAF header
		if { [regexp {^PAF.HDR.START|PAF.HDR.END} $line] } {
		    incr PAFMarkerCount
		}

		# Go to next line
		continue

	    } else {

		# For ephemeris records:

		# remove extra spaces around records
		regsub -all {\ ,} $value "," l1
		regsub -all {,\ } $l1 "," value

		# Make a Tcl list with the values on the line
		set varList [split $value ","]

		# A line must contain at least 6 values
		if { [llength $varList] < 6 } {

		    issueError $ob_id "Ephemeris file (line $lineCount): expecting a minimum of 6 comma-separated values!" $runID

		} else {

		    # For lines with the correct number of values:

		    # Increase number of records found
		    incr recordCount

		    # Put individual values into variables
		    set Date [lindex $varList 0]
		    set JDate [lindex $varList 1]
		    set RA [lindex $varList 2]
		    set DEC [lindex $varList 3]
		    set dRA [lindex $varList 4]
		    set dDEC [lindex $varList 5]
		    set mag [lindex $varList 6]

		    # Check the date format: expecting something like '2012-12-23T12:33:54.123'
		    # Issue error if no match.
		    if { ![regexp {^[0-2][0-9][0-9][0-9]\-(0[0-9]|1[0-2])\-(0[1-9]|[1-2][0-9]|3[0-1])T([0-1][0-9]|2[0-3]):[0-5][0-9]:[0-5]?[0-9](\.[0-9]+)?$} $Date] } {
			issueError $ob_id "Ephemeris file (line $lineCount, $Date): found inconsistent date '$Date'" $runID

		    } else {

			## For valid dates: check overlap with time intervals

			# Convert date to 'clock value' (UTC)
			regsub {\.[0-9]*$} $Date "" tempDate
			regsub {T} $tempDate " " tempDate
			set dateTime [clock scan $tempDate -gmt 1]

			# Check if date matches one of the time intervals (that were not matched before)
			for {set i 0} {$i < [llength $timeIntList] } {incr i} {

			    set ti [lindex $timeIntList $i]
			    set ti_begin [lindex $ti 0] ; set ti_end [lindex $ti 1]

			    # If match found, remove time interval from the list
			    if { $dateTime >= $ti_begin && $dateTime <= $ti_end } {
				set timeIntList [lreplace $timeIntList $i $i]
			    }
			}

			# Check if day is covered by time intervals
			set day [clock format $dateTime -format "%Y-%m-%d"]

			# If same day is found, remove it from the list
			if { [set idx [lsearch -exact $dateList $day]] != -1 } {
			    set dateList [lreplace $dateList $idx $idx]
			}

		    }

		    # Check RA format: expecting something like '16 05 54.123'
		    if { ![regexp {^([0-1][0-9]|2[0-3])\ [0-5][0-9]\ [0-5][0-9](\.[0-9]*)?$} $RA] } {
			issueError $ob_id "Ephemeris file (line $lineCount, $Date): found inconsistent RA '$RA'" $runID
		    } else {
			regsub -all {\ } $RA "" tempRA
			set RAarcsec [ra2arcsec $tempRA]
		    }

		    # Check DEC format: expecting something like '-34 54 44.212'
		    if { ![regexp {^([+\-]?[0-8][0-9]\ [0-5][0-9]\ [0-5][0-9](\.[0-9]*)?|[+\-]?90\ 00\ 00(\.0*)?)$} $DEC ] } {
			issueError $ob_id "Ephemeris file (line $lineCount, $Date): found inconsistent DEC '$DEC'" $runID
		    } else {
			regsub -all {\ } $DEC "" tempDEC
			set DECarcsec [dec2arcsec $tempDEC]
		    }

		    # Check dRA format: expecting '-0.1242323'
		    if { [regexp {^[+\-]?[0-9]+(\.[0-9]*)?$} $dRA] } {

			# Absolute value for apparent motion cannot be more
			# than 1 arcsec/second
			if { [expr abs($dRA)] > 1 } {
			    issueWarning $ob_id "Ephemeris file (line $lineCount, $Date): RA apparent motion is too large: did you use the correct units of arcsecond/second?"
			}

		    } else {
			issueError $ob_id "Ephemeris file (line $lineCount, $Date): found inconsistent RA apparent motion '$dRA'" $runID
		    }

		    # Check dDEC format: same as dRA
		    if { [regexp {^[+\-]?[0-9]+(\.[0-9]*)?$} $dDEC] } {

			# Absolute value for apparent motion cannot be more
			# than 1 arcsec/second
			if { [expr abs($dDEC)] > 1 } {
			    issueWarning $ob_id "Ephemeris file (line $lineCount, $Date): DEC apparent motion is too large: did you use the correct units of arcsecond/second?"
			}

		    } else {
			issueError $ob_id "Ephemeris file (line $lineCount, $Date): found inconsistent DEC apparent motion '$dDEC'" $runID
		    }

		    # Check magnitude format (optional value): expecting '14.6565'
		    if { $mag != "" && ![regexp {^[+\-]?[0-9][0-9]?(\.[0-9]*)?$} $mag] } {

			issueError $ob_id "Ephemeris file (line $lineCount, $Date): found inconsistent magnitude '$mag'" $runID
		    }

		    # Check the angular distance between consecutive ephemeris records,
		    # if they are closer in time than 2 hours: must not exceed 30 arcsec
		    if { $previousDateTime != "" && [expr $dateTime - $previousDateTime] <= 7200 && \
			[set dist [computeDistance $previousRAarcsec $RAarcsec $previousDECarcsec $DECarcsec]] > 30 } {

			issueWarning $ob_id [format "Ephemeris file (line $lineCount, $Date): angular distance w/r to previous entry is larger than 30 arcseconds (found %.1f arcsec)" $dist]

		    }

		    # Record date + RA + DEC for comparison with next line
		    set previousDateTime $dateTime
		    set previousRAarcsec $RAarcsec
		    set previousDECarcsec $DECarcsec

		}

	    }

	}

	# Overall checks

	# If list of time intervals is not empty, this means no overlap was found with
	# the ephemeris file: raise an ERROR.
	foreach ti $timeIntList {

	    set timeStart [clock format [lindex $ti 0] -format "%Y-%m-%dT%H:%M" -gmt 1]
	    set timeEnd [clock format [lindex $ti 1] -format "%Y-%m-%dT%H:%M" -gmt 1]

	    set msg "The following time interval does not match the period covered in the ephemeris file: \[$timeStart - $timeEnd\]"
	    issueError $ob_id $msg $runID

	}

	if { [llength $dateList] > 0 } {

	    set dd [join [uniq $dateList] ", "]
	    set dlist ""

	    set len 100
	    while { [string length $dd] > 0 } {
		set brk [string last " " $dd $len]
		if { $brk < 0 } {set brk $len}
		append dlist "[string range $dd 0 $brk]\\n"
		set dd [string range $dd [expr {$brk+1}] end]
	    }

	    set msg "Please note that the following dates corresponding to your time constraints cannot be found in the ephemeris file:\n[subst $dlist]"
	    issueInfo $ob_id $msg

	}

	# Case of no valid records found (wrong formatting or wrong file attached?)
	if { $recordCount == 0 } {

	    set msg "The ephemeris file has no valid records in it. Please check the file contents."
	    issueError $ob_id $msg $runID

	} elseif { $recordCount > 55000 } {

	    # Case of over-long ephemeris file
	    set msg "Found more than 55000 records in the ephemeris file."
	    issueError $ob_id $msg $runID

	} elseif { $recordCount >= 1 && $PAFMarkerCount < 2 } {

	    # Case of missing PAF header
	    set msg "Found ephemeris records but no valid PAF header. Please check the format of the file."
	    issueWarning $ob_id $msg

	}

    }

}; # End checkEphemerisFile

# Procedure: parseCFFile
# Parses the .cf file and returns the bluest wavelength
# arguments:
#  - ob: handle to the OB object
#
# returns:
#  - wavelengthList: list of all wavelengths in the OB
#
proc parseCFFile { ob } {

    global OBdata
    # set CFFile "/home/smysore/OHS/p2pp-3.4.0-SINGLE-IP/instruments/VIMOS-95.00/VIMOS.cf"
    # Read .cf file. Variable CFData contains the whole file
    # set fp [open $CFFile r]
    # set CFData [read $fp]
    # close $fp

    set CFData [evmInstrumentCfg $ob]

    #puts $CFData

    #
    # Parse the .cf data
    #

    # Array: list of optical elements per template
    array set CFDataTpl {}

    # Array: list of wavelengths per element names
    # May contain values such as CFDataElt(INS.GRIS.WLEN) = WAVELENGTH
    array set CFDataElt {}

    # Array: (fixed) wavelength to consider for some specific templates
    array set CFDataTplFixed {}

    # Split into individual lines
    set CFData [split $CFData "\n"]

    # Only these PAF keywords are of interest
    set CFKeywords { "TPL.NAME" "TPL.WAVE" "TPL.WAVE.UNIT" "TPL.DEFAULT" }

    # Initialize unit: default = nanometer
    set multiplicator 1

    # Loop on all CF lines
    foreach line $CFData {

	# Parse the PAF line
	set PAFContent [parsePAFLine $line]

	# Skip lines that are not lines with PAF content or with non-relevant content
	if { [llength $PAFContent] == 0 || \
	    [lsearch -exact $CFKeywords [set keyword [lindex $PAFContent 0]]] == -1 } {
	    continue
	}

	# Remove spaces and convert CSV string to list
	regsub -all "\[ \t\]" [lindex $PAFContent 1] "" value
	set value [split $value ","]

	# If PAF line contains CF data, parse and store data accordingly
	switch -exact $keyword {

	    "TPL.NAME" {
		set CFDataTpl([lindex $value 0]) [lrange $value 1 end]
	    }

	    "TPL.WAVE" {
		set CFDataElt([lindex $value 0]) [lindex $value 1]
	    }

	    "TPL.WAVE.UNIT" {
		switch $value {
		    "nanometer" { set multiplicator 1 }
		    "micron"    { set multiplicator 1000 }
		}
	    }

	    "TPL.DEFAULT" {
		set CFDataTplFixed([lindex $value 0]) [lindex $value 1]
	    }

	}

    }

    # Post-processing if unit != nanometer: conversion needed
    if { $multiplicator != 1 } {
	foreach elt [array names CFDataElt] {

	    if { $CFDataElt($elt) == "WAVELENGTH" } { continue }
	    set CFDataElt($elt) [expr $CFDataElt($elt) * $multiplicator]

	}
    }

    #parray CFDataTpl
    #parray CFDataElt
    #parray CFDataTplFixed

    #	# 2nd part: construct list of wavelengths based on OB content
    #
    #	## FOR TEST PURPOSES: SIMULATED DATA
    #	set ob_id 123456
    #	array set OBdata {}
    #
    #	set OBdata($ob_id,nbTpl) 3
    #
    #	set OBdata($ob_id,0,tplName) UVES_blue_acq
    #	set OBdata($ob_id,1,tplName) UVES_blue_obs_exp
    #	set OBdata($ob_id,2,tplName) UVES_blue_obs_expfree
    #
    #	set OBdata($ob_id,1,INS.BLUEEXP.MODE) "346+860"
    #	set OBdata($ob_id,2,INS.GRAT1.WLEN) 432
    #
    #	## END TEST

    # Prepare lists of values + wavelengths found in OB
    set wavelengthList {}
    set valueList {}

    set ob_id [evmID $ob]
    # Loop on all OB templates
    for {set i 0} {$i < $OBdata($ob_id,nbTpl)} {incr i} {

	# Get the template name
	set tplName $OBdata($ob_id,$i,tplName)

	# If template corresponds to fixed wavelength, simply append value to list
	if { [info exists CFDataTplFixed($tplName)] } {

	    lappend wavelengthList [expr $CFDataTplFixed($tplName) * $multiplicator]
	    lappend valueList $tplName

	} else {

	    # If template not referenced in .cf file
	    # --> skip to next template
	    if { ![info exists CFDataTpl($tplName)]} { continue }

	    # Loop on optical elements for that template
	    foreach key1 $CFDataTpl($tplName) {
		set key [string trim $key1]
		# If keyword directly referenced in CFDataElt,
		# append (numerical) value of said keyword
		if { [info exists CFDataElt($key)] } {

		    lappend wavelengthList [expr $OBdata($ob_id,$i,$key) * $multiplicator]
		    lappend valueList $key

		} else {
		    # If not, lookup value(s) of keyword in CFDataElt
		    # to get the numerical value. Loop is included in case the
		    # keyword contains a _list_ of values (e.g. several filters)

		    if { [info exists OBdata($ob_id,$i,$key)] } {

			foreach val $OBdata($ob_id,$i,$key) {

			    if { ($val == "none" || $val == "Free") && ![info exists CFDataElt([string trim $val])] } {

				# Default wavelength for settings such as 'none' or 'Free'
				# in case not explicitly referenced in .cf file (failsafe)
				lappend wavelengthList 600
				lappend valueList $val

			    } else {

				lappend wavelengthList $CFDataElt([string trim $val])
				lappend valueList $val

			    }

			}
		    }
		}

	    }
	}
    }

    # Failsafe: in case no wavelengths are found, consider the V-band
    if { [llength $wavelengthList] == 0 } {
	set wavelengthList { 600 }
    }

    #    puts "---PARSE CF FILE---"
    #    puts "valueList=$valueList"
    #    puts "wavelengthList=$wavelengthList"

    # Post-processing: account for case when 'none' or 'Free' value is used

    if { [llength [uniq $valueList]] > 1 && [lsearch -regexp $valueList {^(Free|none)$} ] != -1 } {

	while { [set index [lsearch -regexp $valueList {^(Free|none)$}]] != -1 } {

	    set valueList [lreplace $valueList $index $index]
	    set wavelengthList [lreplace $wavelengthList $index $index]

	}

    }

    #    puts "AFTER POST-PROCESSING (none/Free values):"
    #    puts "valueList=$valueList"
    #    puts "wavelengthList=$wavelengthList"
    #    puts "---END: PARSE CF FILE---"

    return $wavelengthList

}; #End parseCFFile

# Procedure: checkImageQuality
#
# arguments:
#  - ob: handle to the OB object
#  - runID: runID of the OB
# returns:
#  - none: all checks performed here
#
proc checkImageQuality { ob runID } {

    global OBdata Phase1data instr_FK_ITF

    # Get ob_id
    set ob_id [evmID $ob]

    # Instrument Name
    set instName [evmInstrumentName $ob]

    #parse the .cf file and get the list of wavelengths
    set wavelengthList [parseCFFile $ob]

    #sort list in ascending
    set sortedList [lsort -real -increasing $wavelengthList]

    #get the bluest wavelength in the OB
    set lambda [string trim [lindex $sortedList 0]]

    ## DEBUG
    #puts "List of wavelengths in the OB (in nanometers)"
    #puts " $sortedList "
    #puts "\nThe bluest wavelength is $lambda nanometers"

    #	Get the airmass value from the OB's constraint set. If it is larger than
    #	1.5 set it to 1.5 (reasoning: the airmass of observation is not known a
    #	priori; for OBs that have a very relaxed airmass constraint the resulting
    #	IQ value at the airmass limit would be quite high, and probably the users
    #	had put a more stringent airmass constraint in his/her ETC calculations,
    #	which we cannot know).

    set airmass $OBdata($ob_id,csAirmass)
    if { $airmass > 1.5 } {
	set airmass 1.5
    }

    # Compute the image quality
    #	The final image quality is calculated from the 3 components:
    #
    #	1) an atmospheric part, taking into account the wave-front outer-scale L_0,
    #		 the Kolb factor F_Kolb (which depends on telescope diameter and L_0)
    #		 and the Fried parameter at the requested wavelength and airmass, r_0.
    #	2) telescope part: depends on telescope diameter and wavelength
    #	3) instrument part: depends on the instrument transfer function (ITF);
    #		 in our case one number per instrument should be accurate enough

    #FWHM_IQ = sqrt(FWHM_atm2 + FWHM_tel2 + FWHM_ins2)

    #	1. Atmospheric part:
    #	-----------------

    # Diameters: L_0 = 46m,	D(VLT) = 8.2m,	D(VISTA) = 4.0m,	D(VST) = 2.5m
    # and values for the average FWHM_ins in instr_FK_ITF(instrument name) = { instr_name, F_Kolb,  average FWHM_ins }
    # VLT case:	F_Kolb = 1/(1+300*D/L_0)-1 = 1/(1+300*8.2/46)-1 = -0.99
    #   VST case: F_Kolb = -0.97,	VISTA case:	F_Kolb = -0.98

    set L_0 46

    set F_Kolb [lindex $instr_FK_ITF($instName) 1]

    #r_0 = 0.976 * 500 * 10-9 / seeing * ((180/pi)*3600) * (lambda/500)**1.2 * airmass**(-0.6)
    set r_0	 [ expr 0.100657 / $Phase1data(seeing) * pow($lambda/500.0,1.2) * pow($airmass,(-0.6)) ]

    set sqrt_param [ expr 1+ $F_Kolb * 2.182 *  pow($r_0/$L_0,0.356) ]

    if { $sqrt_param < 0 } {
	set FWHM_atm 0
    } else {
	set FWHM_atm [expr $Phase1data(seeing) *  pow( $lambda/500.0,(-0.2)) *  pow($airmass,(0.6)) *  sqrt($sqrt_param) ]
    }

    #	2. Telescope part:
    #	-----------------

    set tel_name [lindex $instr_FK_ITF($instName) 0]

    if { $tel_name == "VLT" } {
	set FWHM_tel [ expr -0.00046 + 0.0000259 * $lambda ]
    } elseif { $tel_name == "VISTA" } {
	set FWHM_tel [ expr 0.00256 + 0.0000558 * $lambda ]
    } elseif  { $tel_name == "VST" } {
	set FWHM_tel [ expr -0.00048 + 0.0000784 * $lambda ]
    } else {
	set FWHM_tel 0
    }

    #	3. Instrument part:
    #	-----------------

    set FWHM_ins [lindex $instr_FK_ITF($instName) 2]

    #Finally, calculate the image quality that should be provided by the user

    set IQ_P1 [ expr sqrt( pow($FWHM_atm,2) +  pow($FWHM_tel,2) +  pow($FWHM_ins,2)) ]

    #perform checks
    #user defined seeing from P2PP
    set IQ_P2  $OBdata($ob_id,csSeeing)

    # Set observing mode (visitor/service)
    set obRunMode $OBdata($ob_id,runMode)

    set print_IQP1 [string range $IQ_P1 0 [ expr {[string first "." $IQ_P1]+1}]]

    # DEBUG
    # puts "Estimated IQ based on Phase 1 seeing: $IQ_P1 arcsec"

    if { [lindex $instr_FK_ITF($instName) 3] != "" } {

	set errorMarginIQ [lindex $instr_FK_ITF($instName) 3]

    } else {

	set errorMarginIQ 0.05

    }

    if { $IQ_P1 < 2.0 && [expr $IQ_P2 + $errorMarginIQ] < $IQ_P1 } {

	set msg "The specified Image Quality = $IQ_P2 is too strict and is not compatible with the Phase 1 seeing value. \n \
				 Please note: based on the V-band seeing at zenith specified at Phase 1, the computed\
				 Image Quality for this OB is [format %.1f $IQ_P1] (arcsec)."
	if { $obRunMode == "s" } {
	    issueError $ob_id $msg $runID
	} else {
	    issueWarning $ob_id $msg
	}
    }

    if { $IQ_P2 > [expr $IQ_P1 + 0.3] } {
	set msg "The specified Image Quality = $IQ_P2 is relaxed with respect to the Phase 1 constraints - is this intentional? \n \
				 Please note: based on the V-band seeing at zenith specified at Phase 1, the computed\
				 Image Quality for this OB is [format %.1f $IQ_P1] (arcsec)."
	issueWarning $ob_id $msg
    }

} ; # End proc checkImageQuality

# Procedure: checkFC_Coordinates
#
# Checks the coherence of the target coordinates between OB and
# attached finding charts (FC). Only works for FC created with
# the Guidecam tool
#
# arguments:
#  - ob_id: ID of the OB
#  - objRef: handle to the OB object
# returns:
#  - none: all checks performed here
#
proc checkFC_Coordinates { ob_id objRef } {

    global OBdata

    # Counter for those FCs that do have metadata from Guidecam
    set FC_Guidecam_count 0

    # Loop on all finding charts
    foreach fc [evmFindingCharts $objRef] {

	# Retrieve the metadata (empty if not present in the FC)
	set meta [evmFCMetadata $fc]

	# Retrieve the target coordinates from the FC
	set RA_FC ""
	set DEC_FC ""
	regexp {"TEL.TARG.ALPHA": *"([0-9:.]+)"} $meta dummy RA_FC
	regexp {"TEL.TARG.DELTA": *"([+-]?[0-9:.]+)"} $meta dummy DEC_FC

	# If coordinates present in the metadata...
	if { $RA_FC != "" && $DEC_FC != "" } {

	    # count this FC
	    incr FC_Guidecam_count

	    # Convert coordinates to arcsec for easier comparison
	    regsub -all {:} $RA_FC "" RA_FC
	    set RA_FC_arcsec [ra2arcsec $RA_FC]
	    regsub -all {:} $DEC_FC "" DEC_FC
	    set DEC_FC_arcsec [dec2arcsec $DEC_FC]

	    # Compare with OB target coordinates (rounded to nearest arcsec), issue warning if different
	    if { [expr round($OBdata($ob_id,ra_marcsec)/1000.)] != [expr round($RA_FC_arcsec)] || \
		[expr round($OBdata($ob_id,dec_marcsec)/1000.)] != [expr round($DEC_FC_arcsec)]} {

		set msg "the finding chart '[evmFCName $fc]' has different coordinates than the ones found in this OB."
		issueWarning $ob_id $msg
	    }
	}
    }

    # Additional info message: how many FCs attached to the OB have been created with Guidecam
    if { $FC_Guidecam_count } {
	set nbFC [llength [evmFindingCharts $objRef]]
	issueInfo $ob_id "$FC_Guidecam_count out of $nbFC finding chart(s) have been created with the Guidecam tool."
    }

} ; # End proc checkFC_Coordinates

######################################################################
##
## Auxiliary procedures for error/warning/info handling and reporting
##
######################################################################

# Global tables where messages are stored
global infoMsg warningMsg errorMsg errorWaiverMsg

array set infoMsg {}
array set warningMsg {}
array set errorMsg {}
array set errorWaiverMsg {}

# issueInfo
#
# Procedure: issue an informational message.
#
# inputs:
#   - id: OB or container ID
#   - message: text of the message to display
#
# output:
#   none. Stores the message into global table
#   for later display
#
proc issueInfo { id message } {

    global infoMsg

    # Find all info messages for the container/OB
    # that are already stored
    set indexes [array names infoMsg "$id,*"]

    # If no info messages found, set the counter to 1
    if { [llength $indexes] == 0 } {

	set infoIdx 1

    } else {

	#...else, get the highest index and add 1
	regexp {[^,]+$} [lindex [lsort -dictionary $indexes] end] lastIndex
	set infoIdx [expr $lastIndex + 1]

    }

    # Store the info message in global table
    set infoMsg($id,$infoIdx) $message

}

# issueWarning
#
# Procedure: issue a warning message.
#
# inputs:
#   - id: OB or container ID
#   - message: text of the message to display
#
# output:
#   none. Stores the message into global table
#   for later display
#
proc issueWarning { id message } {

    global warningMsg

    # Find all warning messages for the container/OB
    # that are already stored
    set indexes [array names warningMsg "$id,*"]

    # If no warning messages found, set the counter to 1
    if { [llength $indexes] == 0 } {

	set infoIdx 1

    } else {

	#...else, get the highest index and add 1
	regexp {[^,]+$} [lindex [lsort -dictionary $indexes] end] lastIndex
	set infoIdx [expr $lastIndex + 1]

    }

    # Store the warning message in global table
    set warningMsg($id,$infoIdx) $message

}

# issueError
#
# Procedure: issue an error  message.
#
# inputs:
#   - id: OB or container ID
#   - message: text of the message to display
#   - runID: ID of the observing run
#
# output:
#   none. Stores the message into global table
#   for later display
#
proc issueError { id message runID } {

    global errorMsg

    # Find all error messages for the container/OB
    # that are already stored
    set indexes [array names errorMsg "$id,*"]

    # If no error messages found, set the counter to 1
    if { [llength $indexes] == 0 } {

	set infoIdx 1

    } else {

	#...else, get the highest index and add 1
	regexp {[^,]+$} [lindex [lsort -dictionary $indexes] end] lastIndex
	set infoIdx [expr $lastIndex + 1]

    }

    set errorIndex 0

    # Store the error message in global table
    set errorMsg($id,$infoIdx) [list $message $runID $errorIndex]

}

# issueErrorWaiver
#
# Procedure: issue a message for a waiverable error.
#
# inputs:
#   - id: OB or container ID
#   - message: text of the message to display
#   - runID: ID of the observing run
#   - errorIndex: indexing number for the waiver condition
#   - advertise: if set to false then we do not show the error
#     as being a waiverable one in the report (default = true).
#
# output:
#   none. Stores the message into global table
#   for later display
#
proc issueErrorWaiver { id message runID errorIndex {advertise "true"} } {

    global errorWaiverMsg

    # Find all messages for waiverable errors for the container/OB
    # that are already stored
    set indexes [array names errorWaiverMsg "$id,*"]

    # If no such messages found, set the counter to 1
    if { [llength $indexes] == 0 } {

	set infoIdx 1

    } else {

	#...else, get the highest index and add 1
	regexp {[^,]+$} [lindex [lsort -dictionary $indexes] end] lastIndex
	set infoIdx [expr $lastIndex + 1]

    }

    # Store the message in global table
    # NB: we store a list with the message, the run ID and the index number
    set errorWaiverMsg($id,$infoIdx) [list $message $runID $errorIndex $advertise]

}

# checkForWaiver
#
# Searches for a possible waiver granted to a run
#
# Inputs: runID (id of the observing run)
#         errorIndex (id of the error/waiver condition)
#
proc checkForWaiver { runID errorIndex } {

    # global array storing the list of waivers granted
    # to observing runs. This array is defined in a
    # separate file, filled out by USG support astronomers,
    # and copied to the P2PP Application Server.
    # This file is NOT a part of this software module
    global evmWaiver

    # Clean the run ID
    set id [string trim $runID]

    # If an entry for this run exists in the waiver array,
    if { [info exists evmWaiver($id)] } {

	# and if the index of the error/waiver condition is
	# listed in the corresponding value of the array,
	if { [lsearch $evmWaiver($id) $errorIndex] != -1 } {

	    # set flag to 'true'
	    set waiver 1

	    # else,
	} else {

	    # set flag to 'false'
	    set waiver 0

	};#end if

	# else,
    } else {

	# set flag to false.
	set waiver 0

    };#end if

    return $waiver

}

# evmERROR
#
#
#
# This procedure is used in place of the EVM function
# evmError. It enables to take into account possible waivers
# granted for some error conditions.
#
# Inputs:
#   - runID: id of the observing run
#   - objID: local ID of the container or OB
#   - isContainer: set to true if the object is a container
#   - msg: error message to display
#   - errorIndex: index number of the error/waiver condition
#   - showWaiverMsg: if set to false then we do not show the error
#     as being a waiverable one in the report (default = true).
#
# Remark: passing 9999 as the value for errorIndex activates the
# so-called 'super-waiver' mechanism for the concerned observing run.
# In such a case, all errors automatically become warnings, so that
# any OB/container is allowed to be checked-in regardless of its
# contents (they always pass verification).
#
# Remark: the EVM API for P2PP 3.x uses distinct functions to raise
# errors for containers and OBs.
#
# Returns:
#  -1 if the 'super-waiver' has been activated
#   0 if this is a normal error
#   1 if a waiver has been granted
#
proc evmERROR { runID objID isContainer msg errorIndex {showWaiverMsg "true"}} {

    global evmWaiver

    set retFlag 0

    # If "super-waiver" has been granted
    if { [checkForWaiver $runID 9999] } {

	# Issue a warning
	evmWarning $msg
	set retFlag -1

	# If the error index is zero,
    } elseif { $errorIndex == 0 } {

	# then issue a normal error message
	if { $isContainer } {
	    evmContainerError $objID $msg
	} else {
	    evmObError $objID $msg
	}

	# else,
    } else {

	# If a waiver has been granted,
	if { [checkForWaiver $runID $errorIndex] } {

	    set retFlag 1

	    # Issue a simple warning, and the info
	    # that a waiver has been granted.
	    if { $showWaiverMsg == "true" } {
		evmWarning "$msg\n$evmWaiver(wavMsg)"
	    } else {
		evmWarning "$msg"
	    }

	    # else,
	} else {

	    # Issue an error
	    if { $showWaiverMsg == "true" } {

		set msg "$msg\n$evmWaiver(errMsg)"

	    }

	    if { $isContainer } {
		evmContainerError $objID $msg
	    } else {
		evmObError $objID $msg
	    }

	} ;# end if

    } ; # end if

    return $retFlag
}

# reportAllMessages
#
# Procedure: displays all info, warning and error messages
# for a given OB or container
#
# inputs:
#   - id: local ID of container/OB
#   - isContainer: must be set to true if the object is a container
#     (default = false)
#
# outputs:
#   none. Displays messages in the EVM report of P2PP
#
proc reportAllMessages { id {isContainer 0} } {

    # Global tables containing stored messages
    global infoMsg warningMsg errorMsg errorWaiverMsg

    global OBdata

    # Step 1: Display all error messages
    set errorList [lsort -dictionary [array names errorMsg "$id,*"]]

    if { [set nbErr [llength $errorList]] > 0 } {

	# Report total number of errors for the container/OB
	if {$nbErr>1} {set p S} else {set p ""}
	evmInfo "--> $nbErr ERROR$p found: must be fixed."

	# Go through the list of errors and issue messages
	foreach errIdx $errorList {

	    #	    if { $isContainer } {
	    #		evmContainerError $id $errorMsg($errIdx)
	    #	    } else {
	    #		evmObError $id $errorMsg($errIdx)
	    #	    }

	    set msg [lindex $errorMsg($errIdx) 0]
	    set runID [lindex $errorMsg($errIdx) 1]
	    set errorIndex [lindex $errorMsg($errIdx) 2]

	    evmERROR $runID $id $isContainer $msg $errorIndex

	}

	evmInfo ""
    }

    # Step 2: Display all messages for "waiverable" errors
    set errorWavList [lsort -dictionary [array names errorWaiverMsg "$id,*"]]

    if { [set nbErrWav [llength $errorWavList]] > 0 } {

	# Report total number of waiverable errors
	if {$nbErrWav>1} {set p S} else {set p ""}
	evmInfo "--> $nbErrWav WAIVERABLE ERROR$p found: must be authorized."

	# Go through the list and issue messages
	foreach errIdx $errorWavList {

	    set msg [lindex $errorWaiverMsg($errIdx) 0]
	    set runID [lindex $errorWaiverMsg($errIdx) 1]
	    set errorIndex [lindex $errorWaiverMsg($errIdx) 2]

	    evmERROR $runID $id $isContainer $msg $errorIndex

	}

	evmInfo ""
    }

    # Step 3: Display all warning messages
    set warningList [lsort -dictionary [array names warningMsg "$id,*"]]

    if { [set nbWarn [llength $warningList]] > 0 } {

	# Report total number of warnings for the container/OB
	if {$nbWarn>1} {set p S} else {set p ""}
	evmInfo "--> $nbWarn WARNING$p: please check."

	# Go through the list and issue messages
	foreach warnIdx $warningList {
	    evmWarning $warningMsg($warnIdx)
	}

	evmInfo ""
    }

    # Step 4: Display all info messages
    set infoList [lsort -dictionary [array names infoMsg "$id,*"]]

    if { [set nbInfo [llength $infoList]] > 0 } {

	# Report total number of info messages
	if {$nbInfo>1} {set p s} else {set p ""}
	evmInfo "--> $nbInfo info message$p"

	# Go through the list and issue messages
	foreach infoIdx $infoList {
	    evmInfo $infoMsg($infoIdx)
	}

	evmInfo ""
    }

    # If no messages at all were found for the container/OB, simply
    # print an "OK" message.
    if { [expr $nbErr + $nbErrWav + $nbWarn + $nbInfo] == 0 } {

	evmInfo "all OK.\n"

    }
}

# getNbWarnings
#
# Procedure: returns how many warnings there are for a given
#            container or OB
#
# inputs:
#   - id: local ID of container or OB
#
# outputs:
#   returns number of warnings
#
proc getNbWarnings { id } {

    global warningMsg

    set warningList [lsort -dictionary [array names warningMsg "$id,*"]]
    set nbWarn [llength $warningList]

    return $nbWarn

}

# getNbErrors
#
# Procedure: returns how many errors (waiverable or not) there are for a given
#            container or OB
#
# inputs:
#   - id: local ID of container or OB
#
# outputs:
#   returns number of errors + waiverable errors
#
proc getNbErrors { id } {

    global errorMsg errorWaiverMsg

    set errorList [lsort -dictionary [array names errorMsg "$id,*"]]
    set errorWavList [lsort -dictionary [array names errorWaiverMsg "$id,*"]]
    set nbErr [expr [llength $errorList] + [llength $errorWavList]]

    return $nbErr

}

# hasMessages
#
# Procedure: determines whether a container or OB has messages
#            (info, warnings or errors) associated to it
#
# inputs:
#   - id: local ID of container or OB
#
# outputs:
#   returns 'true' if messages were found, 'false' if not
#
proc hasMessages { id } {

    global infoMsg warningMsg errorMsg errorWaiverMsg

    # Get number of errors for the object
    set errorList [lsort -dictionary [array names errorMsg "$id,*"]]
    set nbErr [llength $errorList]

    # Get number of waiverable errors for the object
    set errorWavList [lsort -dictionary [array names errorWaiverMsg "$id,*"]]
    set nbErrWav [llength $errorWavList]

    # Get number of warnings for the object
    set warningList [lsort -dictionary [array names warningMsg "$id,*"]]
    set nbWarn [llength $warningList]

    # Get number of infos for the object
    set infoList [lsort -dictionary [array names infoMsg "$id,*"]]
    set nbInfo [llength $infoList]

    # Sum everything, then return true or false depending on the result
    if { [expr $nbErr + $nbErrWav + $nbWarn + $nbInfo] > 0 } {
	return true
    } else {
	return false
    }

}

# displayEVMSummaryTable
#
# Procedure: displays a summary table inside the EVM report, showing
# numbers of warnings and errors for all selected OBs and containers,
# as well as a global OK/FAILED status
#
# inputs:
#   none
#
# outputs:
#   none. Info is displayed inside the EVM report directly
#
proc displayEVMSummaryTable {} {

    # Global tables holding information about warnings, errors, etc.
    global infoMsg warningMsg errorMsg errorWaiverMsg
    global compoundList OBdata

    # Table header
    evmInfo "OB VALIDATION: SUMMARY\n"
    evmInfo "+===================================+==========+========+========+
|Object name                        | Warnings | Errors | Status |
+===================================+==========+========+========+"

    # Go through each object part of the selection
    foreach object $compoundList {

	# Get 'object type' (first element of object list)
	set objType [lindex $object 0]

	if { [llength $objType] > 1 } {

	    # Case of Container
	    set containerInfo [lindex $object 0]

	    set containerID [lindex $containerInfo 1] ; # Container ID
	    set objRef  [lindex $containerInfo 2] ; # Object reference (Java handle)
	    set objSelected [lindex $containerInfo 3] ; # Is container selected or not
	    set containerName [string range [evmName $objRef] 0 34]

	    # List of OBs in the container
	    set listOfObsBlocks [lindex $object 1]

	    # Is the container explicitly selected?
	    if { $objSelected } {

		# ... then get numbers of warnings and errors for
		# this container
		set nbw [getNbWarnings $containerID]
		set nbe [getNbErrors $containerID]

		# If errors found, status = FAILED...
		if { $nbe > 0 } {
		    set status FAILED
		} else {

		    # If not, status = OK...
		    set status OK

		    # .. but also look at the status of the OBs!

		    # Loop on all OBs in the container: if one has errors,
		    # then the status of the whole container becomes FAILED
		    foreach ob $listOfObsBlocks {
			if { ![lindex $ob 3] } { continue }
			set ob_id [lindex $ob 1]
			set nbeOB [getNbErrors $ob_id]
			if { $nbeOB > 0 } {
			    set status FAILED
			    break
			}
		    }
		}

	    } else {

		# If the container was not explicitly selected, do not report on it
		set nbw "--"
		set nbe "--"
		set status "--"

	    }

	    # Print line of the table for that container
	    set line [format "|%-35s| % 8s | % 6s | %6s |" \
		$containerName $nbw $nbe $status]
	    evmInfo $line

	    # Then go through the list of OBs of the container...

	    # Loop on the OBs
	    foreach ob $listOfObsBlocks {

		# If the OB was not explicitly selected, skip it
		if { [lindex $ob 3] } { set sel YES } else { set sel NO }
		if { $sel == "NO" } { continue }

		# Get OB ID
		set ob_id [lindex $ob 1]

		# Get OB name (limit to 30 characters for formatting)
		set OBname [string range $OBdata($ob_id,name) 0 30]

		# Get numbers of warnings + errors
		set nbw [getNbWarnings $ob_id]
		set nbe [getNbErrors $ob_id]

		# Determine global status (has errors or not)
		if { $nbe > 0 } {
		    set status FAILED
		} else {
		    set status OK
		}

		# Print line of the table for this OB
		set line [format "|  * %-31s| % 8s | % 6s | %6s |" \
		    $OBname $nbw $nbe $status]
		evmInfo $line

	    }

	} else {

	    # Case of Single OB
	    set objSelected [lindex $object 3] ; # Is object part of selection or not
	    set ob_id [lindex $object 1] ; # OB ID
	    set OBname [string range $OBdata($ob_id,name) 0 34]

	    # We only report on selected OBs (should always be the case for loose OBs)
	    if { $objSelected } {

		# Get number of warnings + errors
		set nbw [getNbWarnings $ob_id]
		set nbe [getNbErrors $ob_id]

		# Determine global status (has errors or not)
		if { $nbe > 0 } {
		    set status FAILED
		} else {
		    set status OK
		}

		# Print line of the table for this OB
		set line [format "|%-35s| % 8s | % 6s | %6s |" \
		    $OBname $nbw $nbe $status]
		evmInfo $line

	    }

	}

	# Separator line
	evmInfo "+-----------------------------------+----------+--------+--------+"
    }

}

# displayEVMReport
#
# Procedure: display a full OB validation report for all
# selected Obs and containers.
#
# inputs:
#   none
#
# outputs:
#   none. Messages are issued in the EVM report
#
proc displayEVMReport {} {

    # Global tables holding messages for warnings, errors, etc.
    global infoMsg warningMsg errorMsg errorWaiverMsg
    global compoundList OBdata

    evmInfo "OB VALIDATION: COMPLETE REPORT\n"

    # Go through the compound list, produce report
    foreach object $compoundList {

	# Get 'object type' (first element of object list)
	set objType [lindex $object 0]

	set listOfObsBlocks {}
	set isContainer false

	# If the object type is a list, then we are dealing
	# with a container: the actual type has to be fetched
	# deeper inside the object list
	if { [llength $objType] > 1 } {

	    set containerInfo [lindex $object 0]

	    set objType [lindex $containerInfo 0] ; # Type of container
	    set containerID [lindex $containerInfo 1] ; # Container ID
	    set objRef  [lindex $containerInfo 2] ; # Object reference (Java handle)
	    set objSelected [lindex $containerInfo 3] ; # Is container selected or not
	    set containerName [evmName $objRef]

	    set listOfObsBlocks [lindex $object 1]
	    set isContainer true

	} else {

	    set objRef [lindex $object 2] ; # Object reference (Java handle)
	    set objSelected [lindex $object 3] ; # Is object part of selection or not
	    set ob_id [lindex $object 1] ; # OB ID
	    set listOfObsBlocks [list $object]
	}

	# Case of single OB
	if { !$isContainer } {

	    # If there are messages linked to this OB, display a header + all messages
	    if { [hasMessages $ob_id] } {
		evmInfo "--==="
		evmInfo "--=== Checks for OB $ob_id ($OBdata($ob_id,name)) ===---"
		evmInfo "--===\n"

		reportAllMessages $ob_id
	    }

	} else {

	    # First: check if there are any messages for the container or its OBs
	    # If not, don't do anything.
	    set needAction false
	    set needActionOB false

	    # Check to see if there are messages associated to this container
	    if { [hasMessages $containerID] } {
		set needAction true
	    }

	    # Go through list of OBs, check if some have messages
	    foreach ob $listOfObsBlocks {
		set ob_id [lindex $ob 1]
		if { [hasMessages $ob_id] } {
		    set needAction true
		    set needActionOB true
		    break
		}
	    }

	    if { $needAction } {

		# Case of Container
		evmInfo "--==="
		evmInfo "--=== Checks for $objType '$containerName' ===---"
		evmInfo "--===\n"

		if { [hasMessages $containerID] } {

		    # 1. Deal with all container-level messages
		    evmInfo "==> Checks at container-level"

		    reportAllMessages $containerID 1

		}

		if { $needActionOB } {

		    # 2. Go through list of OBs
		    evmInfo "==> Checks of individual OBs"

		    foreach ob $listOfObsBlocks {

			set ob_id [lindex $ob 1]

			# If OB has messages, display header + all messages
			if { [hasMessages $ob_id] } {
			    evmInfo "+--> OB $ob_id ($OBdata($ob_id,name))"
			    reportAllMessages $ob_id
			}

		    }
		}
	    }
	}

    }

}

